module 'data.array.backtrackable' [Impl] {
   new(Metrics, InitialValue, Array) :-
      Impl:compute_metrics(Metrics, Offsets, Sizes),
      'kernel':array_backtrackable_new(Sizes, InitialValue, X),
      Impl:array(Offsets, X, Array).

   init(Metrics, F, Array) :-
      Impl:compute_metrics(Metrics, Offsets, Sizes),
      'kernel':array_backtrackable_new(Sizes, 0, X),
      module Iter [Iter, F, X] {
         iter([], Concrete, Abstract) :-
            'kernel':array_backtrackable_set(X, Concrete, F:get(Abstract)).

         iter([range(Off, Low, Up, _, _) | T], Concrete, Abstract) :-
            'data.number':iter(Low, Up) [Iter, T, Concrete, Abstract, Off] {
               do(I) :-
                  Iter:iter(
                     T,
                     ['data.number':(Off + I) | Concrete],
                     [I | Abstract]
                  ).
            }
      }:iter('data.list':reverse(Offsets), [], []),
      Impl:array(Offsets, X, Array).

   new(Metrics) = 'data.array.backtrackable':init(Metrics) {
      get(_) = _.
   }.
},

module Impl {
   compute_metrics(Metrics, Offsets, Sizes) :-
      'data.list':iter(Metrics, Offsets, Sizes) {
         do(Range, Offset, Size) :-
            (
               Range = Lower .. Upper
            ->
               Max = 'data.number':(Upper - Lower),
               Offset = range('data.number':(-Lower), Lower, Upper, 0, Max),
               Size = 'data.number':succ(Max)
            ;
               Max = 'data.number':pred(Range),
               Offset = range(0, 0, Max, 0, Max),
               Size = Range
            ).
      }.

   compute_indexes(Offsets, Index0, Index1) :-
      module Aux [Aux, Index0] {
         do([], _, []).
         do([H | T], I0, I1) :-
            (
               H = range(Off, Low, High, _, _)
            ->
               I0 = [I | I0T],
               (
                  'data.number':(
                     Low =< I,
                     I =< High
                  )
               ->
                  true
               ;
                  throw(error(index_out_of_bounds(Index0), array))
               ),
               I1 = ['data.number':(Off + I) | I1T],
               Aux:do(T, I0T, I1T)
            ;
               H = fixed(I)
            ->
               I1 = [I | I1T],
               Aux:do(T, I0, I1T)
            ).
      },
      Aux:do(Offsets, Index0, Index1).

   array(Offsets, X, Array) :-
      Sizes = 'data.list':morph(Offsets) {
         do(range(_, Lower, Upper, _, _)) = [Lower .. Upper].
         do(fixed(_)) = [].
      },
      Dim = 'data.list':length(Sizes),
      module Array [Dim, Offsets, Sizes, Array, X] {
         [Index | Tail] = Value :-
            Array:get([Index | Tail], Value).

         get(Indexes, Value) :-
            (
               'data.list':length(Indexes, Dim)
            ->
               compute_indexes(Offsets, Indexes, Indexes0),
               'kernel':array_backtrackable_get(X, Indexes0, Value)
            ;
               throw(error(dimension_mismatch, array))
            ).

         set(Indexes, Value) :-
            (
               'data.list':length(Indexes, Dim)
            ->
               compute_indexes(Offsets, Indexes, Indexes0),
               'kernel':array_backtrackable_set(X, Indexes0, Value)
            ;
               throw(error(dimension_mismatch, array))
            ).

         sizes(Sizes).

         slice(Metrics, NewSlice) :-
            (
               'data.list':length(Metrics, Dim)
            ->
               true
            ;
               throw(error(dimension_mismatch, array))
            ),
            module Aux [Aux, Metrics] {
               compute_dim(Low, Up, Lower, Upper, NewLow, NewUp) :-
                  (
                     'data.fd':set_domain(Lower, Low, Up),
                     'data.fd':set_domain(Upper, Low, Up),
                     'data.fd.fac':(Lower =< Upper),
                     NewUp = 'data.fd.fac':(NewLow + Upper - Lower)
                  ->
                     true
                  ;
                     throw(error(invalid_bounds(Metrics), array))
                  ),
                  'data.fd':labeling([Lower, Upper]),
                  (
                     NewLow = Low
                  ->
                     true
                  ;
                     true
                  ).

               do([], _, []).
               do([Offset | Offsets], List, NewOffsets) :-
                  (
                     Offset = range(Off, Low, Up, _, _)
                  ->
                     List = [H | T],
                     (
                        (
                           H = (Old -> Target)
                        ->
                           Target = NewLow .. NewUp
                        ;
                           H = Old
                        ),                           
                        Old = Lower .. Upper
                     ->
                        Aux:compute_dim(Low, Up, Lower, Upper, NewLow, NewUp),
                        CLow = 'data.number':(Off + Lower),
                        CUp = 'data.number':(Off + Upper),
                        Off0 = 'data.number':(CLow - NewLow),
                        NewOffset = range(Off0, NewLow, NewUp, CLow, CUp),
                        NewOffsets = [NewOffset | TailOffsets],
                        Aux:do(Offsets, T, TailOffsets)
                     ;
                        (
                           'data.number':(
                              Low =< H,
                              H =< Up
                           )
                        ->
                           true
                        ;
                           throw(error(index_out_of_bounds, array))
                        ),
                        NewOffset = fixed('data.number':(Off + H)),
                        NewOffsets = [NewOffset | TailOffsets],
                        Aux:do(Offsets, T, TailOffsets)
                     )
                  ;
                     NewOffsets = [Offset | TailOffsets],
                     Aux:do(Offsets, T, TailOffsets)
                  ).
            },
            Aux:do(Offsets, Metrics, NewOffsets),
            array(NewOffsets, X, NewSlice).

         list = Iter:iter(Offsets, []) :-
            module Iter [X, Iter] {
               iter([], Indexes) = [Value] :-
                  'kernel':array_backtrackable_get(X, Indexes, Value).

               iter([Offset | Tail], Indexes, List) :-
                  (
                     Offset = range(_, _, _, Low, High)
                  ->
                     List =
                        'data.list':flatten(
                           'data.list':init(Low, High) [
                              Iter, Indexes, Tail
                           ] {
                              get(Index) =
                                 Iter:iter(
                                    Tail,
                                    'data.list':append(Indexes, [Index])
                                 ).
                           }
                        )
                  ;
                     Offset = fixed(Index)
                  ->
                     Iter:iter(
                        Tail,
                        'data.list':append(Indexes, [Index]),
                        List
                     )
                  ).
            }.

         deep_list = Iter:iter(Offsets, []) :-
            module Iter [X, Iter] {
               iter([], Indexes) = Value :-
                  'kernel':array_backtrackable_get(X, Indexes, Value).

               iter([Offset | Tail], Indexes, List) :-
                  (
                     Offset = range(_, _, _, Low, High)
                  ->
                     List =
                        'data.list':init(Low, High) [
                           Iter, Indexes, Tail
                        ] {
                           get(Index) =
                              Iter:iter(
                                 Tail,
                                 'data.list':append(Indexes, [Index])
                              ).
                        }
                  ;
                     Offset = fixed(Index)
                  ->
                     Iter:iter(
                        Tail,
                        'data.list':append(Indexes, [Index]),
                        List
                     )
                  ).
            }.

         pretty = 'data.array.backtrackable:'(Array:deep_list)
      }.
      
}